@@ -21,15 +21,14 @@
  * Boston, MA 02111-1307, USA.
  */
 
-#include <libxfce4util/libxfce4util.h>
-
 #include "util.h"
 #include "file.h"
 #include "thumbnailer.h"
-#include "settings.h"
 #include "icon_bar.h"
 #include "main_window.h"
 
+#include <glib/gi18n.h>
+
 
 
 enum
@@ -94,6 +93,9 @@ static gboolean
 rstto_icon_bar_draw (GtkWidget *widget,
                      cairo_t *cr);
 static gboolean
+rstto_icon_bar_enter (GtkWidget *widget,
+                      GdkEventCrossing *event);
+static gboolean
 rstto_icon_bar_leave (GtkWidget *widget,
                       GdkEventCrossing *event);
 static gboolean
@@ -141,6 +143,9 @@ cb_rstto_thumbnail_size_changed (GObject *settings,
                                  GParamSpec *pspec,
                                  gpointer user_data);
 static void
+cb_rstto_icon_bar_adjustment_changed (GtkAdjustment *adjustment,
+                                      RsttoIconBar *icon_bar);
+static void
 rstto_icon_bar_update_missing_icon (RsttoIconBar *icon_bar);
 
 
@@ -244,6 +249,7 @@ rstto_icon_bar_class_init (RsttoIconBarClass *klass)
     gtkwidget_class->get_preferred_height = rstto_icon_bar_get_preferred_height;
     gtkwidget_class->size_allocate = rstto_icon_bar_size_allocate;
     gtkwidget_class->draw = rstto_icon_bar_draw;
+    gtkwidget_class->enter_notify_event = rstto_icon_bar_enter;
     gtkwidget_class->leave_notify_event = rstto_icon_bar_leave;
     gtkwidget_class->motion_notify_event = rstto_icon_bar_motion;
     gtkwidget_class->button_press_event = rstto_icon_bar_button_press;
@@ -568,6 +574,10 @@ rstto_icon_bar_set_property (
                                       G_CALLBACK (gtk_widget_queue_draw), icon_bar);
             g_signal_connect_swapped (icon_bar->priv->vadjustment, "value-changed",
                                       G_CALLBACK (gtk_widget_queue_draw), icon_bar);
+            g_signal_connect (icon_bar->priv->hadjustment, "changed",
+                              G_CALLBACK (cb_rstto_icon_bar_adjustment_changed), icon_bar);
+            g_signal_connect (icon_bar->priv->vadjustment, "changed",
+                              G_CALLBACK (cb_rstto_icon_bar_adjustment_changed), icon_bar);
             break;
 
         default:
@@ -631,6 +641,7 @@ rstto_icon_bar_realize (GtkWidget *widget)
     attributes.height = MAX (icon_bar->priv->height, allocation.height);
     attributes.event_mask = (GDK_SCROLL_MASK
             | GDK_EXPOSURE_MASK
+            | GDK_ENTER_NOTIFY_MASK
             | GDK_LEAVE_NOTIFY_MASK
             | GDK_POINTER_MOTION_MASK
             | GDK_BUTTON_PRESS_MASK
@@ -728,9 +739,7 @@ rstto_icon_bar_size_allocate (GtkWidget *widget,
                               GtkAllocation *allocation)
 {
     RsttoIconBar *icon_bar = RSTTO_ICON_BAR (widget);
-    GtkAllocation prev_alloc;
 
-    gtk_widget_get_allocation (widget, &prev_alloc);
     gtk_widget_set_allocation (widget, allocation);
 
     if (!icon_bar->priv->active_item)
@@ -742,10 +751,6 @@ rstto_icon_bar_size_allocate (GtkWidget *widget,
                                 allocation->width, allocation->height);
         gdk_window_resize (icon_bar->priv->bin_window, allocation->width, allocation->height);
     }
-
-    /* center on the active item when there is a real size change */
-    if (prev_alloc.width != allocation->width || prev_alloc.height != allocation->height)
-        rstto_icon_bar_show_active (icon_bar);
 }
 
 
@@ -788,18 +793,29 @@ rstto_icon_bar_draw (GtkWidget *widget,
 
 
 
+static void
+rstto_icon_bar_update_cursor_item (RsttoIconBar *icon_bar,
+                                   RsttoIconBarItem *item)
+{
+    if (icon_bar->priv->cursor_item != NULL)
+        rstto_icon_bar_queue_draw_item (icon_bar, icon_bar->priv->cursor_item);
+
+    icon_bar->priv->cursor_item = item;
+    if (icon_bar->priv->cursor_item != NULL)
+        rstto_icon_bar_queue_draw_item (icon_bar, icon_bar->priv->cursor_item);
+}
+
+
+
 static gboolean
-rstto_icon_bar_leave (
-        GtkWidget        *widget,
-        GdkEventCrossing *event)
+rstto_icon_bar_enter (GtkWidget *widget,
+                      GdkEventCrossing *event)
 {
     RsttoIconBar *icon_bar = RSTTO_ICON_BAR (widget);
+    RsttoIconBarItem *item;
 
-    if (icon_bar->priv->cursor_item != NULL)
-    {
-        rstto_icon_bar_queue_draw_item (icon_bar, icon_bar->priv->cursor_item);
-        icon_bar->priv->cursor_item = NULL;
-    }
+    item = rstto_icon_bar_get_item_at_pos (icon_bar, event->x, event->y);
+    rstto_icon_bar_update_cursor_item (icon_bar, item);
 
     return FALSE;
 }
@@ -807,51 +823,37 @@ rstto_icon_bar_leave (
 
 
 static gboolean
-rstto_icon_bar_motion (
-        GtkWidget      *widget,
-        GdkEventMotion *event)
+rstto_icon_bar_leave (GtkWidget *widget,
+                      GdkEventCrossing *event)
 {
+    rstto_icon_bar_update_cursor_item (RSTTO_ICON_BAR (widget), NULL);
+
+    return FALSE;
+}
+
+
+
+static gboolean
+rstto_icon_bar_motion (GtkWidget *widget,
+                       GdkEventMotion *event)
+{
+    RsttoIconBar *icon_bar = RSTTO_ICON_BAR (widget);
     RsttoIconBarItem *item;
-    RsttoIconBar     *icon_bar = RSTTO_ICON_BAR (widget);
-    GtkTreeIter       iter;
-    RsttoFile        *file;
 
     item = rstto_icon_bar_get_item_at_pos (icon_bar, event->x, event->y);
-    if (item != NULL && icon_bar->priv->cursor_item != item)
-    {
-        if (icon_bar->priv->cursor_item != NULL)
-            rstto_icon_bar_queue_draw_item (icon_bar, icon_bar->priv->cursor_item);
-        icon_bar->priv->cursor_item = item;
-        rstto_icon_bar_queue_draw_item (icon_bar, item);
+    rstto_icon_bar_update_cursor_item (icon_bar, item);
 
-        iter = item->iter;
-        gtk_tree_model_get (icon_bar->priv->model, &iter,
-                icon_bar->priv->file_column, &file,
-                -1);
-        g_object_unref (file);
-
-        gtk_widget_trigger_tooltip_query (widget);
-    }
-    else if (icon_bar->priv->cursor_item != NULL
-            && icon_bar->priv->cursor_item != item)
-    {
-        rstto_icon_bar_queue_draw_item (icon_bar, icon_bar->priv->cursor_item);
-        icon_bar->priv->cursor_item = NULL;
-    }
-
-    return TRUE;
+    return FALSE;
 }
 
 static gboolean
-rstto_icon_bar_scroll (
-        GtkWidget      *widget,
-        GdkEventScroll *event)
+rstto_icon_bar_scroll (GtkWidget *widget,
+                       GdkEventScroll *event)
 {
-    RsttoIconBar  *icon_bar   = RSTTO_ICON_BAR (widget);
-    GtkAdjustment *adjustment = NULL;
-    gdouble        val        = 0;
-    gdouble        step_size  = 0;
-    gdouble        max_value  = 0;
+    RsttoIconBar *icon_bar = RSTTO_ICON_BAR (widget);
+    RsttoIconBarItem *item;
+    GtkAdjustment *adjustment;
+    gdouble val, old_val, step_size, max_value;
 
     if (icon_bar->priv->orientation == GTK_ORIENTATION_VERTICAL)
         adjustment = icon_bar->priv->vadjustment;
@@ -859,7 +861,7 @@ rstto_icon_bar_scroll (
         adjustment = icon_bar->priv->hadjustment;
 
     step_size = icon_bar->priv->item_size / 2.0;
-    val = gtk_adjustment_get_value (adjustment);
+    val = old_val = gtk_adjustment_get_value (adjustment);
     max_value = gtk_adjustment_get_upper (adjustment) - gtk_adjustment_get_page_size (adjustment);
 
     switch (event->direction)
@@ -890,6 +892,12 @@ rstto_icon_bar_scroll (
     }
 
     gtk_adjustment_set_value (adjustment, val);
+    if (icon_bar->priv->orientation == GTK_ORIENTATION_VERTICAL)
+        item = rstto_icon_bar_get_item_at_pos (icon_bar, event->x, event->y + (val - old_val));
+    else
+        item = rstto_icon_bar_get_item_at_pos (icon_bar, event->x + (val - old_val), event->y);
+
+    rstto_icon_bar_update_cursor_item (icon_bar, item);
 
     return TRUE;
 }
@@ -1897,6 +1905,33 @@ cb_rstto_thumbnail_size_changed (
     rstto_icon_bar_update_missing_icon (icon_bar);
 }
 
+static void
+cb_rstto_icon_bar_adjustment_changed (GtkAdjustment *adjustment,
+                                      RsttoIconBar *icon_bar)
+{
+    gdouble page_size, upper;
+
+    static gdouble s_page_size = 0, s_upper = 0;
+
+    if ((icon_bar->priv->orientation == GTK_ORIENTATION_VERTICAL
+         && adjustment != icon_bar->priv->vadjustment)
+        || (icon_bar->priv->orientation == GTK_ORIENTATION_HORIZONTAL
+            && adjustment != icon_bar->priv->hadjustment))
+        return;
+
+    page_size = gtk_adjustment_get_page_size (adjustment);
+    upper = gtk_adjustment_get_upper (adjustment);
+
+    /* center on the active item when there is a real size change: with the allocation mode
+     * applied in get_preferred_width/height(), 'page_size' corresponds to the visible part
+     * of the icon bar, while 'upper' corresponds to its allocated size */
+    if (s_page_size != page_size || s_upper != upper)
+    {
+        s_page_size = page_size;
+        s_upper = upper;
+        rstto_icon_bar_show_active (icon_bar);
+    }
+}
 
 static void
 rstto_icon_bar_update_missing_icon (RsttoIconBar *icon_bar)
