@@ -20,20 +20,22 @@
  *    Copyright (c) Benedict Meurer <benny@xfce.org>
  */
 
-#include <string.h>
-
 #include "util.h"
 #include "image_list.h"
 #include "thumbnailer.h"
-#include "settings.h"
 #include "main_window.h"
 
+#include <glib/gi18n.h>
+
 
 #define FILE_BLOCK_SIZE 10000
+#define ERROR_LOAD_DIR_NO_VALID_FILE _("No supported image type found in the directory")
+#define ERROR_LOAD_DIR_PARTIAL _("Directory only partially loaded")
+#define ERROR_LOAD_DIR_FAILED _("Could not load directory")
+#define ERROR_INVALID_FILE _("Unsupported mime type")
 
 enum
 {
-    RSTTO_IMAGE_LIST_SIGNAL_REMOVE_IMAGE = 0,
     RSTTO_IMAGE_LIST_SIGNAL_REMOVE_ALL,
     RSTTO_IMAGE_LIST_SIGNAL_SORTED,
     RSTTO_IMAGE_LIST_SIGNAL_COUNT
@@ -182,7 +184,6 @@ struct _RsttoImageListPrivate
 
     GList        *image_monitors;
     GList        *images;
-    gint          n_images;
 
     GSList       *iterators;
     GCompareFunc  cb_rstto_image_list_compare_func;
@@ -262,18 +263,6 @@ rstto_image_list_class_init (RsttoImageListClass *klass)
 
     object_class->finalize = rstto_image_list_finalize;
 
-    rstto_image_list_signals[RSTTO_IMAGE_LIST_SIGNAL_REMOVE_IMAGE] = g_signal_new ("remove-image",
-            G_TYPE_FROM_CLASS (klass),
-            G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
-            0,
-            NULL,
-            NULL,
-            g_cclosure_marshal_VOID__OBJECT,
-            G_TYPE_NONE,
-            1,
-            G_TYPE_OBJECT,
-            NULL);
-
     rstto_image_list_signals[RSTTO_IMAGE_LIST_SIGNAL_REMOVE_ALL] = g_signal_new ("remove-all",
             G_TYPE_FROM_CLASS (klass),
             G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
@@ -343,84 +332,53 @@ rstto_image_list_new (void)
 }
 
 gboolean
-rstto_image_list_add_file (
-        RsttoImageList *image_list,
-        RsttoFile *r_file,
-        GError **error)
+rstto_image_list_add_file (RsttoImageList *image_list,
+                           RsttoFile *r_file,
+                           GError **error)
 {
-    GList *image_iter = g_list_find (image_list->priv->images, r_file);
-    GSList *iter = image_list->priv->iterators;
-    gint i = 0;
-    GtkTreePath *path = NULL;
+    GtkTreePath *path;
     GtkTreeIter t_iter;
-    GFileMonitor *monitor = NULL;
+    GFileMonitor *monitor;
+    GSList *iter;
+    gint idx = 0;
 
-    g_return_val_if_fail (NULL != r_file, FALSE);
     g_return_val_if_fail (RSTTO_IS_FILE (r_file), FALSE);
 
-    if (!image_iter)
+    /* file already added */
+    if (g_list_find (image_list->priv->images, r_file))
+        return TRUE;
+
+    /* unsupported mime type */
+    if (! rstto_file_is_valid (r_file))
     {
-        if (r_file)
-        {
-            if (rstto_file_is_valid (r_file))
-            {
-                g_object_ref (r_file);
-
-                image_list->priv->images = g_list_insert_sorted (
-                        image_list->priv->images,
-                        r_file,
-                        rstto_image_list_get_compare_func (image_list));
-
-                image_list->priv->n_images++;
-
-                if (image_list->priv->dir_monitor == NULL)
-                {
-                    monitor = g_file_monitor_file (
-                            rstto_file_get_file (r_file),
-                            G_FILE_MONITOR_NONE,
-                            NULL,
-                            NULL);
-                    g_signal_connect (monitor, "changed",
-                                      G_CALLBACK (cb_file_monitor_changed), image_list);
-                    image_list->priv->image_monitors = g_list_prepend (
-                            image_list->priv->image_monitors,
-                            monitor);
-                }
-                i = g_list_index (image_list->priv->images, r_file);
-
-                path = gtk_tree_path_new ();
-                gtk_tree_path_append_index (path, i);
-                t_iter.stamp = image_list->priv->stamp;
-                t_iter.user_data = r_file;
-                t_iter.user_data3 = GINT_TO_POINTER (i);
-
-                gtk_tree_model_row_inserted (
-                        GTK_TREE_MODEL (image_list),
-                        path,
-                        &t_iter);
-
-                gtk_tree_path_free (path);
-
-                /** TODO: update all iterators */
-                while (iter)
-                {
-                    if (! RSTTO_IMAGE_LIST_ITER (iter->data)->priv->sticky)
-                    {
-                        rstto_image_list_iter_find_file (iter->data, r_file);
-                    }
-                    iter = g_slist_next (iter);
-                }
-
-                return TRUE;
-            }
-            else
-            {
-                return FALSE;
-            }
-        }
+        g_set_error (error, G_IO_ERROR, G_IO_ERROR_INVALID_ARGUMENT, ERROR_INVALID_FILE);
         return FALSE;
     }
 
+    image_list->priv->images = g_list_insert_sorted (image_list->priv->images, g_object_ref (r_file),
+                                                     image_list->priv->cb_rstto_image_list_compare_func);
+
+    if (image_list->priv->dir_monitor == NULL)
+    {
+        monitor = g_file_monitor_file (rstto_file_get_file (r_file), G_FILE_MONITOR_NONE, NULL, NULL);
+        g_signal_connect (monitor, "changed", G_CALLBACK (cb_file_monitor_changed), image_list);
+        image_list->priv->image_monitors = g_list_prepend (image_list->priv->image_monitors, monitor);
+    }
+
+    path = gtk_tree_path_new ();
+    idx = g_list_index (image_list->priv->images, r_file);
+    gtk_tree_path_append_index (path, idx);
+    t_iter.stamp = image_list->priv->stamp;
+    t_iter.user_data = r_file;
+    t_iter.user_data3 = GINT_TO_POINTER (idx);
+
+    gtk_tree_model_row_inserted (GTK_TREE_MODEL (image_list), path, &t_iter);
+    gtk_tree_path_free (path);
+
+    for (iter = image_list->priv->iterators; iter != NULL; iter = iter->next)
+        if (! RSTTO_IMAGE_LIST_ITER (iter->data)->priv->sticky)
+            rstto_image_list_iter_find_file (iter->data, r_file);
+
     return TRUE;
 }
 
@@ -476,7 +434,7 @@ rstto_image_list_remove_file (
         {
             if (rstto_file_equal (rstto_image_list_iter_get_file (iter->data), r_file))
             {
-                if (rstto_image_list_iter_get_position (iter->data) == n_images -1)
+                if (rstto_image_list_iter_get_position (iter->data) == n_images - 1)
                 {
                     iter_previous (iter->data, FALSE);
                 }
@@ -524,10 +482,6 @@ rstto_image_list_remove_file (
             iter = g_slist_next (iter);
         }
 
-        g_signal_emit (image_list,
-                       rstto_image_list_signals[RSTTO_IMAGE_LIST_SIGNAL_REMOVE_IMAGE],
-                       0, r_file, NULL);
-
         g_object_unref (r_file);
         gtk_tree_path_free (path_);
     }
@@ -552,18 +506,101 @@ rstto_image_list_remove_all (RsttoImageList *image_list)
                    0, NULL);
 }
 
+static gboolean
+rstto_image_list_set_directory_finish_idle (gpointer data)
+{
+    RsttoImageList *image_list;
+    RsttoFile *r_file;
+    GtkFileFilter *filter;
+    GtkFileFilterInfo filter_info;
+    GFileEnumerator *file_enum = data;
+    GList *info_list, *li;
+    GFile *file, *loaded_file = NULL;
+
+    filter = rstto_main_window_get_app_file_filter ();
+    filter_info.contains = GTK_FILE_FILTER_MIME_TYPE;
+
+    image_list = rstto_object_get_data (file_enum, "image-list");
+    info_list = rstto_object_get_data (file_enum, "info-list");
+    r_file = rstto_object_get_data (file_enum, "loaded-file");
+
+    if (r_file != NULL)
+        loaded_file = rstto_file_get_file (r_file);
+
+    for (li = info_list; li != NULL; li = li->next)
+    {
+        filter_info.mime_type = g_file_info_get_content_type (li->data);
+        if (filter_info.mime_type != NULL && gtk_file_filter_filter (filter, &filter_info))
+        {
+            /* skip already loaded file, if any */
+            file = g_file_enumerator_get_child (file_enum, li->data);
+            if (loaded_file != NULL && g_file_equal (file, loaded_file))
+            {
+                loaded_file = NULL;
+                rstto_object_set_data (file_enum, "loaded-file", NULL);
+                g_object_unref (file);
+
+                continue;
+            }
+
+            /* load the first valid file, if needed */
+            r_file = rstto_file_new (file);
+            g_object_unref (file);
+            if (loaded_file == NULL && image_list->priv->images == NULL)
+            {
+                if (! rstto_image_list_add_file (image_list, r_file, NULL))
+                {
+                    g_object_unref (r_file);
+                    continue;
+                }
+                else
+                {
+                    /* give the viewer a chance to display the image first */
+                    rstto_object_set_data (file_enum, "loaded-file", r_file);
+                    g_idle_add (rstto_image_list_set_directory_finish_idle, file_enum);
+                    g_object_unref (r_file);
+
+                    return FALSE;
+                }
+            }
+
+            /* no additional filtering here, it will be done gradually when validating files,
+             * and the mime type will be refined at this time as well */
+            rstto_file_set_content_type (r_file, filter_info.mime_type);
+            image_list->priv->images = g_list_prepend (image_list->priv->images, r_file);
+        }
+    }
+
+    /* schedule next file block, if any */
+    if (g_list_length (info_list) == FILE_BLOCK_SIZE)
+        g_idle_add (rstto_image_list_set_directory_idle, file_enum);
+    else
+    {
+        /* we're done, inform the others */
+        image_list->priv->is_busy = FALSE;
+        if (image_list->priv->images != NULL)
+            rstto_image_list_set_compare_func (image_list,
+                                               image_list->priv->cb_rstto_image_list_compare_func);
+        else
+            rstto_util_dialog_error (ERROR_LOAD_DIR_NO_VALID_FILE, NULL);
+
+        g_object_unref (file_enum);
+    }
+
+    g_list_free_full (info_list, g_object_unref);
+
+    return FALSE;
+}
+
 static void
 rstto_image_list_set_directory_finish (GObject *source_object,
                                        GAsyncResult *res,
                                        gpointer user_data)
 {
     RsttoImageList *image_list = user_data;
-    RsttoFile *r_file;
     GFileEnumerator *file_enum = G_FILE_ENUMERATOR (source_object);
-    GList *info_list, *li;
-    GFile *file, *loaded_file = NULL;
+    GList *info_list;
     GError *error = NULL;
-    const gchar *content_type;
 
     if (rstto_main_window_get_app_exited ())
     {
@@ -576,60 +613,26 @@ rstto_image_list_set_directory_finish (GObject *source_object,
     {
         /* we're done, inform the others */
         image_list->priv->is_busy = FALSE;
-        rstto_image_list_set_compare_func (image_list,
-                                           image_list->priv->cb_rstto_image_list_compare_func);
-
-        if (error != NULL)
+        if (image_list->priv->images != NULL)
         {
-            /* TODO: show error dialog */
-            g_error_free (error);
+            rstto_image_list_set_compare_func (image_list,
+                                               image_list->priv->cb_rstto_image_list_compare_func);
+            if (error != NULL)
+                rstto_util_dialog_error (ERROR_LOAD_DIR_PARTIAL, error);
         }
+        else
+            rstto_util_dialog_error (ERROR_LOAD_DIR_NO_VALID_FILE, error);
 
         g_object_unref (file_enum);
+        if (error != NULL)
+            g_error_free (error);
 
         return;
     }
 
-    r_file = rstto_object_get_data (file_enum, "loaded-file");
-    if (r_file != NULL)
-        loaded_file = rstto_file_get_file (r_file);
-
-    for (li = info_list; li != NULL; li = li->next)
-    {
-        content_type  = g_file_info_get_content_type (li->data);
-        if (content_type != NULL && g_str_has_prefix (content_type, "image/"))
-        {
-            file = g_file_enumerator_get_child (file_enum, li->data);
-            if (loaded_file != NULL && g_file_equal (file, loaded_file))
-            {
-                loaded_file = NULL;
-                rstto_object_set_data (file_enum, "loaded-file", NULL);
-                g_object_unref (file);
-
-                continue;
-            }
-
-            /* no filtering here, it will be done when requesting thumbnails */
-            image_list->priv->images = g_list_prepend (image_list->priv->images,
-                                                       rstto_file_new (file));
-            g_object_unref (file);
-        }
-    }
-
-    /* schedule next file block, if any */
-    if (g_list_length (info_list) == FILE_BLOCK_SIZE)
-        g_idle_add (rstto_image_list_set_directory_idle, file_enum);
-    else
-    {
-        /* we're done, inform the others */
-        image_list->priv->is_busy = FALSE;
-        rstto_image_list_set_compare_func (image_list,
-                                           image_list->priv->cb_rstto_image_list_compare_func);
-
-        g_object_unref (file_enum);
-    }
-
-    g_list_free_full (info_list, g_object_unref);
+    /* try to prioritize image display over directory loading */
+    rstto_object_set_data (file_enum, "info-list", info_list);
+    g_idle_add (rstto_image_list_set_directory_finish_idle, file_enum);
 }
 
 static gboolean
@@ -642,14 +645,38 @@ rstto_image_list_set_directory_idle (gpointer data)
     return FALSE;
 }
 
+static void
+rstto_image_list_set_directory_enumerate_finish (GObject *dir,
+                                                 GAsyncResult *res,
+                                                 gpointer user_data)
+{
+    RsttoImageList *image_list = user_data;
+    GFileEnumerator *file_enum;
+    GError *error = NULL;
+
+    /* this is transfer full but this ref will be released in
+     * rstto_image_list_set_directory_finish() when everything is done */
+    file_enum = g_file_enumerate_children_finish (G_FILE (dir), res, &error);
+    if (file_enum == NULL)
+    {
+        image_list->priv->is_busy = FALSE;
+        rstto_util_dialog_error (ERROR_LOAD_DIR_FAILED, error);
+        g_error_free (error);
+
+        return;
+    }
+
+    rstto_object_set_data (file_enum, "loaded-file", rstto_object_get_data (dir, "loaded-file"));
+    rstto_object_set_data (file_enum, "image-list", image_list);
+    g_idle_add (rstto_image_list_set_directory_idle, rstto_util_source_autoremove (file_enum));
+}
+
 gboolean
 rstto_image_list_set_directory (RsttoImageList *image_list,
                                 GFile *dir,
                                 RsttoFile *file,
                                 GError **error)
 {
-    GFileEnumerator *file_enum;
-
     rstto_image_list_remove_all (image_list);
     rstto_image_list_monitor_dir (image_list, dir);
 
@@ -660,17 +687,17 @@ rstto_image_list_set_directory (RsttoImageList *image_list,
     if (file != NULL && ! rstto_image_list_add_file (image_list, file, error))
         return FALSE;
 
-    /* this is transfer full but this ref will be released in
-     * rstto_image_list_set_directory_finish() when everything is done */
-    file_enum = g_file_enumerate_children (dir, "standard::content-type",
-                                           G_FILE_QUERY_INFO_NONE, NULL, error);
-    if (file_enum == NULL)
-        return FALSE;
-
+    /*
+     * Using async methods both to get the dir enum and the file info is needed in general,
+     * as well as specifying "standard::name" in order to be able to use
+     * g_file_enumerator_get_child() afterwards.
+     * See https://gitlab.gnome.org/GNOME/glib/-/issues/2507
+     */
     image_list->priv->is_busy = TRUE;
-    rstto_object_set_data (file_enum, "loaded-file", file);
-    rstto_object_set_data (file_enum, "image-list", image_list);
-    g_idle_add (rstto_image_list_set_directory_idle, rstto_util_source_autoremove (file_enum));
+    rstto_object_set_data (dir, "loaded-file", file);
+    g_file_enumerate_children_async (dir, "standard::name,standard::content-type",
+                                     G_FILE_QUERY_INFO_NONE, G_PRIORITY_DEFAULT, NULL,
+                                     rstto_image_list_set_directory_enumerate_finish, image_list);
 
     return TRUE;
 }
@@ -906,6 +933,7 @@ iter_set_position (
                    rstto_image_list_iter_signals[RSTTO_IMAGE_LIST_ITER_SIGNAL_PREPARE_CHANGE],
                    0, NULL);
 
+    iter->priv->sticky = sticky;
     if (iter->priv->r_file)
     {
         iter->priv->r_file = NULL;
@@ -1014,7 +1042,7 @@ rstto_image_list_iter_has_next (RsttoImageListIter *iter)
     else
     {
         if (rstto_image_list_iter_get_position (iter) ==
-            (rstto_image_list_get_n_images (image_list) -1))
+            (rstto_image_list_get_n_images (image_list) - 1))
         {
             return FALSE;
         }
@@ -1051,12 +1079,14 @@ iter_previous (
     if (position)
     {
         iter->priv->r_file = position->data;
+        ret_val = TRUE;
     }
     else
     {
         if (image_list->priv->wrap_images)
         {
             position = g_list_last (iter->priv->image_list->priv->images);
+            ret_val = TRUE;
         }
         else
         {
@@ -1116,15 +1146,7 @@ rstto_image_list_iter_has_previous (RsttoImageListIter *iter)
 RsttoImageListIter *
 rstto_image_list_iter_clone (RsttoImageListIter *iter)
 {
-    RsttoImageListIter *new_iter = rstto_image_list_iter_new (
-            iter->priv->image_list,
-            iter->priv->r_file);
-
-    rstto_image_list_iter_set_position (
-            new_iter,
-            rstto_image_list_iter_get_position (iter));
-
-    return new_iter;
+    return rstto_image_list_iter_new (iter->priv->image_list, iter->priv->r_file);
 }
 
 GCompareFunc
