@@ -17,27 +17,27 @@
  *  02110-1301, USA.
  */
 
-#include <math.h>
-
-#include <gio/gio.h>
-
 #include "util.h"
 #include "image_viewer.h"
-#include "settings.h"
 #include "main_window.h"
 
+#include <math.h>
 
 
-/* Do not make this buffer too large,
- * this breaks some pixbufloaders.
+
+/*
+ * A buffer size of 1 MiB makes it possible to load a lot of current images in only
+ * a few iterations, which can significantly improve performance for some formats
+ * like GIF, without being too big, which on the contrary would degrade performance.
+ * On the other hand, it ensures that a cancelled load will end fairly quickly at
+ * the end of the current iteration, instead of continuing almost indefinitely as
+ * can be the case for some images, again in GIF format.
+ * See https://gitlab.xfce.org/apps/ristretto/-/issues/16 for an example of such an
+ * image.
  */
-#ifndef RSTTO_IMAGE_VIEWER_BUFFER_SIZE
-#define RSTTO_IMAGE_VIEWER_BUFFER_SIZE 4096
-#endif
+#define LOADER_BUFFER_SIZE 1048576
 
-#ifndef BACKGROUND_ICON_SIZE
 #define BACKGROUND_ICON_SIZE 128
-#endif
 
 #define MAX_OVER_VISIBLE 1.5
 #define MIN_VIEW_PERCENT 0.1
@@ -218,6 +218,7 @@ struct _RsttoImageViewerTransaction
     /* File I/O data */
     /*****************/
     guchar           *buffer;
+    gssize            read_bytes;
 };
 
 struct _RsttoImageViewerPrivate
@@ -238,6 +239,7 @@ struct _RsttoImageViewerPrivate
 
     RsttoImageViewerTransaction *transaction;
     RsttoImageOrientation        orientation;
+    gchar                      **excluded_mime_types;
     struct
     {
         cairo_pattern_t *pattern;
@@ -313,6 +315,9 @@ rstto_image_viewer_post_init (RsttoImageViewer *viewer)
 static void
 rstto_image_viewer_init (RsttoImageViewer *viewer)
 {
+    GSList *list, *li;
+    gchar **strv;
+
     if (default_screen == NULL)
     {
         default_screen = gdk_screen_get_default ();
@@ -374,6 +379,32 @@ rstto_image_viewer_init (RsttoImageViewer *viewer)
     gtk_drag_dest_set (GTK_WIDGET (viewer), GTK_DEST_DEFAULT_ALL, NULL, 0,
                       GDK_ACTION_COPY | GDK_ACTION_LINK | GDK_ACTION_MOVE | GDK_ACTION_PRIVATE);
     gtk_drag_dest_add_uri_targets (GTK_WIDGET (viewer));
+
+    /*
+     * Mime types for which we will use the default pixbuf loader, instead of the
+     * mime-type-specific one, typically because it is broken. This is different from using
+     * gdk_pixbuf_format_set_disabled(), which would totally disable the corresponding library,
+     * whereas some of its features could still be used here.
+     * This should apply to as few mime types as possible, but all RAW types should be excluded
+     * for now, see https://gitlab.freedesktop.org/libopenraw/libopenraw/-/issues/7
+    */
+    viewer->priv->excluded_mime_types = NULL;
+    list = gdk_pixbuf_get_formats ();
+    for (li = list; li != NULL; li = li->next)
+    {
+        /* not sure we can count on the stability of the format name or description, so better
+         * to look for a mime type that we know must be present */
+        strv = gdk_pixbuf_format_get_mime_types (li->data);
+        if (g_strv_contains ((const gchar *const *) strv, "image/x-canon-cr2"))
+        {
+            viewer->priv->excluded_mime_types = strv;
+            break;
+        }
+        else
+            g_strfreev (strv);
+    }
+
+    g_slist_free (list);
 }
 
 /**
@@ -598,6 +629,16 @@ rstto_image_viewer_finalize (GObject *object)
         g_object_unref (viewer->priv->iter);
         viewer->priv->iter = NULL;
     }
+    if (viewer->priv->excluded_mime_types)
+    {
+        g_strfreev (viewer->priv->excluded_mime_types);
+        viewer->priv->excluded_mime_types = NULL;
+    }
+    if (viewer->priv->file)
+    {
+        g_object_unref (viewer->priv->file);
+        viewer->priv->file = NULL;
+    }
 
     G_OBJECT_CLASS (rstto_image_viewer_parent_class)->finalize (object);
 }
@@ -1322,6 +1363,7 @@ static void
 rstto_image_viewer_load_image (RsttoImageViewer *viewer, RsttoFile *file, gdouble scale)
 {
     RsttoImageViewerTransaction *transaction = g_new0 (RsttoImageViewerTransaction, 1);
+    const gchar *mime_type;
 
     /*
      * This will first need to return to the 'main' loop before it cleans up after itself.
@@ -1333,16 +1375,17 @@ rstto_image_viewer_load_image (RsttoImageViewer *viewer, RsttoFile *file, gdoubl
         viewer->priv->transaction = NULL;
     }
 
-    transaction->loader = gdk_pixbuf_loader_new_with_mime_type (rstto_file_get_content_type (file), NULL);
+    mime_type = rstto_file_get_content_type (file);
+    if (viewer->priv->excluded_mime_types == NULL
+        || ! g_strv_contains ((const gchar *const *) viewer->priv->excluded_mime_types, mime_type))
+        transaction->loader = gdk_pixbuf_loader_new_with_mime_type (mime_type, NULL);
 
-    /* HACK HACK HACK */
+    /* fallback in case of an excluded mime type, or any error */
     if (transaction->loader == NULL)
-    {
         transaction->loader = gdk_pixbuf_loader_new ();
-    }
 
     transaction->cancellable = g_cancellable_new ();
-    transaction->buffer = g_new0 (guchar, RSTTO_IMAGE_VIEWER_BUFFER_SIZE);
+    transaction->buffer = g_new0 (guchar, LOADER_BUFFER_SIZE);
     transaction->file = file;
     transaction->viewer = viewer;
     transaction->scale = scale;
@@ -1354,7 +1397,7 @@ rstto_image_viewer_load_image (RsttoImageViewer *viewer, RsttoFile *file, gdoubl
     viewer->priv->transaction = transaction;
 
     g_file_read_async (rstto_file_get_file (transaction->file),
-                       0,
+                       G_PRIORITY_DEFAULT,
                        transaction->cancellable,
                        cb_rstto_image_viewer_read_file_ready,
                        transaction);
@@ -1517,13 +1560,16 @@ cb_rstto_image_viewer_read_file_ready (GObject *source_object, GAsyncResult *res
     if (rstto_main_window_get_app_exited ())
         return;
 
-    file_input_stream = g_file_read_finish (file, result, NULL);
+    file_input_stream = g_file_read_finish (file, result, &transaction->error);
     if (file_input_stream == NULL)
+    {
+        gdk_pixbuf_loader_close (transaction->loader, NULL);
         return;
+    }
 
     g_input_stream_read_async (G_INPUT_STREAM (file_input_stream),
                                transaction->buffer,
-                               RSTTO_IMAGE_VIEWER_BUFFER_SIZE,
+                               LOADER_BUFFER_SIZE,
                                G_PRIORITY_DEFAULT,
                                transaction->cancellable,
                                cb_rstto_image_viewer_read_input_stream_ready,
@@ -1531,49 +1577,51 @@ cb_rstto_image_viewer_read_file_ready (GObject *source_object, GAsyncResult *res
 }
 
 static void
-cb_rstto_image_viewer_read_input_stream_ready (GObject *source_object, GAsyncResult *result, gpointer user_data)
+loader_write_async (GTask *task,
+                    gpointer source_object,
+                    gpointer task_data,
+                    GCancellable *cancellable)
+{
+    RsttoImageViewerTransaction *transaction = task_data;
+
+    if (! gdk_pixbuf_loader_write (transaction->loader, transaction->buffer,
+                                   transaction->read_bytes, &transaction->error)
+        || g_cancellable_is_cancelled (transaction->cancellable))
+    {
+        gdk_pixbuf_loader_close (transaction->loader, NULL);
+        g_object_unref (source_object);
+    }
+    else
+        g_input_stream_read_async (source_object, transaction->buffer, LOADER_BUFFER_SIZE,
+                                   G_PRIORITY_DEFAULT, transaction->cancellable,
+                                   cb_rstto_image_viewer_read_input_stream_ready, transaction);
+}
+
+static void
+cb_rstto_image_viewer_read_input_stream_ready (GObject *source_object,
+                                               GAsyncResult *result,
+                                               gpointer user_data)
 {
     RsttoImageViewerTransaction *transaction = user_data;
-    gssize read_bytes;
+    GTask *task;
 
     if (rstto_main_window_get_app_exited ())
         return;
 
-    read_bytes = g_input_stream_read_finish (G_INPUT_STREAM (source_object), result, &transaction->error);
-    if (read_bytes == -1)
+    transaction->read_bytes = g_input_stream_read_finish (G_INPUT_STREAM (source_object),
+                                                          result, &transaction->error);
+    if (transaction->read_bytes <= 0)
     {
-        gdk_pixbuf_loader_close (transaction->loader, NULL);
-        return;
-    }
-
-    if (read_bytes > 0)
-    {
-        if (! gdk_pixbuf_loader_write (transaction->loader, transaction->buffer,
-                                       read_bytes, &transaction->error))
-        {
-            /* Clean up the input-stream */
-            g_input_stream_close (G_INPUT_STREAM (source_object), NULL, NULL);
-            g_object_unref (source_object);
-        }
-        else
-        {
-            g_input_stream_read_async (G_INPUT_STREAM (source_object),
-                                       transaction->buffer,
-                                       RSTTO_IMAGE_VIEWER_BUFFER_SIZE,
-                                       G_PRIORITY_DEFAULT,
-                                       transaction->cancellable,
-                                       cb_rstto_image_viewer_read_input_stream_ready,
-                                       transaction);
-        }
+        gdk_pixbuf_loader_close (transaction->loader,
+                                 transaction->read_bytes == 0 ? &transaction->error : NULL);
+        g_object_unref (source_object);
     }
     else
     {
-        /* Loading complete, transaction should not be free-ed */
-        gdk_pixbuf_loader_close (transaction->loader, &transaction->error);
-
-        /* Clean up the input-stream */
-        g_input_stream_close (G_INPUT_STREAM (source_object), NULL, NULL);
-        g_object_unref (source_object);
+        task = g_task_new (source_object, transaction->cancellable, NULL, NULL);
+        g_task_set_task_data (task, transaction, NULL);
+        g_task_run_in_thread (task, loader_write_async);
+        g_object_unref (task);
     }
 }
 
@@ -1696,9 +1744,11 @@ cb_rstto_image_loader_closed_idle (gpointer data)
 
     if (viewer->priv->transaction == transaction)
     {
-        if (transaction->error == NULL
-            || g_error_matches (transaction->error, GDK_PIXBUF_ERROR,
-                                GDK_PIXBUF_ERROR_CORRUPT_IMAGE))
+        if (transaction->error == NULL || (
+                g_error_matches (transaction->error, GDK_PIXBUF_ERROR,
+                                 GDK_PIXBUF_ERROR_CORRUPT_IMAGE)
+                && gdk_pixbuf_loader_get_pixbuf (transaction->loader) != NULL
+            ))
         {
             gtk_widget_set_tooltip_text (widget, NULL);
             cb_rstto_image_loader_image_ready (transaction->loader, transaction);
@@ -1773,6 +1823,8 @@ cb_rstto_image_viewer_update_pixbuf (gpointer user_data)
         viewer->priv->pixbuf.has_alpha = gdk_pixbuf_get_has_alpha (pixbuf);
         viewer->priv->pixbuf.width = gdk_pixbuf_get_width (pixbuf);
         viewer->priv->pixbuf.height = gdk_pixbuf_get_height (pixbuf);
+        set_device_scale (G_OBJECT (gtk_widget_get_ancestor (user_data, RSTTO_TYPE_MAIN_WINDOW)),
+                          NULL, viewer);
 
         /* redraw only the image */
         rect.x = viewer->priv->rendering.x_offset;
