@@ -48,6 +48,9 @@ struct _RsttoMainWindow
 
 
 
+RsttoMainWindow *
+rstto_main_window_get_app_window (void);
+
 gboolean
 rstto_main_window_get_app_exited (void);
 
@@ -62,7 +65,8 @@ RsttoImageListIter *
 rstto_main_window_get_iter (RsttoMainWindow *window);
 
 void
-rstto_main_window_add_file_to_recent_files (GFile *file);
+rstto_main_window_open (RsttoMainWindow *window,
+                        GSList *files);
 
 gboolean
 rstto_main_window_play_slideshow (RsttoMainWindow *window);
